# Self-uprighting & line-following, balancing LEGO Robots
# based on https://legostudiovives.be/balancing_lego_robots/
from pybricks.hubs import InventorHub
from pybricks.pupdevices import Motor, ColorSensor, UltrasonicSensor
from pybricks.parameters import Port, Direction
from pybricks.tools import wait, StopWatch
from pybricks.geometry import Axis

# Initialize motors.
left = Motor(Port.A, Direction.COUNTERCLOCKWISE)
right = Motor(Port.C)
left.reset_angle(0)
right.reset_angle(0)

# Initialize hub and color sensor.
# You can also use the TechnicHub and/or the ColorDistanceSensor instead.
hub = InventorHub()
#BAD hub = PrimeHub()

sensorL = ColorSensor(Port.D)
sensorR = ColorSensor(Port.E)
#eyes = UltrasonicSensor(Port.E)
#eyes.lights.on(10)

# Initialize position buffer for speed calculation.
DT = 10
DRIVE_SPEED = 250
window = int(250/DT)
buf = [0] * window
idx = 0
angle = 0

PAUSE = 5000

# Timer to generate reference position.
watch = StopWatch()

while True:
    # Get angular rate and estimated angle.
    rate = hub.imu.angular_velocity(Axis.X)
    angle += rate*DT/1000

    # Get motor position.
    position = (left.angle() + right.angle()) / 2

    # Calculate motor speed.
    speed = (position - buf[idx]) / (window*DT) * 1000
    buf[idx] = position
    idx = (idx + 1) % window

    # Calculate reference position, which just grows linearly with time.
    reference = max(watch.time() - PAUSE, 0) / 1000 * DRIVE_SPEED

    # Calculate duty cycle.
    #duty = 0.018*rate + 15*angle + 0.2*(position - reference) + 0.16*speed
    duty = 0.018*rate + 15*angle + 0.2*(position - reference) + 0.16*speed

    # Account for battery level and type.
    duty *= 7200/hub.battery.voltage()

    # Calculate steering
    #   for TWO color sensor
    reflectionL = sensorL.reflection()
    reflectionR = sensorR.reflection()
    err = reflectionL - reflectionR
    steering = err*0.45

    #   for SINGLE color sensor
    #err = reflectionL - 45
    #steering = err*0.45

    # Apply duty cycle for balancing and steering.
    left.dc(duty + steering)
    right.dc(duty - steering)

    # Wait some time.
    wait(DT)